/*
 * Scriptographer script  objects-on-paths with jitter and fade 
 *  and fit-randomiser ( GW provate version only ) 
 * @ http://scriptographer.org 
 *
 * Repeat objects or groups along one or more paths
 * Optionally rotate or wiggle the objects
 * Repeat the created path to create some texture like elements.
 * You can also scale the objects during the repetition.
 *
 * Based on script object-on-path 
 * by Luca esnho Moroni esnhoesnho@gmail.com - www.lucamoroni.com
 *
 * Mods by Gareth Williams gw@westdale.demon.co.uk to support
 * multiple Paths and Objects with scaling of spacing and random/cycled drawing
 * 
 * 5 Aug 2012 version -- add scaling of strokewidth as item.scale() does not change it
 */

script.coordinateSystem = 'bottom-up';

// OPTIONS
// preference for fading so last on is not invisible 
var fadeLimit = 0.025;
// preference for randomised fitting 
var ranFitMin = 0.5;
var ranFitMax = 2.0;

var objList;             // the selected objects
var pathList;            // the selected paths

// dialogue text constants
var rotaTypeOff = 'off';
var rotaTypeRotate = 'rotate';
var rotaTypeVib = 'vibra';
var spaceTypeCount = 'count';
var spaceTypeSpace = 'fixed';
var spaceTypeFit = 'fit';
var spaceTypeScaledFit = 'scale';
var spaceTypeFitRand = 'fit-rand';
var spaceTypeScaledFitRand = 'scale-rand';
var repeatTypeAll = 'all';
var repeatTypeRand = 'random';
var repeatTypeCycle = 'cycle';
var growUp = 'increase';
var growDown = 'decrease';
var growOff = 'off'

// cache for parameters with initial values for dialogue
var spacing = spaceTypeCount;
var spaceNo = 20;
var spaceAdjust = 100;
var repeatType = repeatTypeAll;
var fade = false;

var mainScale = 100;
var growAmount = 95;
var grow  = growDown; 
var frontToBack = true; 

var rota = rotaTypeVib;
var vibraAm = 45;
var rotationIncrement = 360/spaceNo;
var rotaPx = 50; // default centre
var rotaPy = 50;
var jitterAmount = 100;

var repeatGroup = false;
var repeatGroupN = 10;
var xOffset = 50;  // later might use object width or percent of path width
var yOffset = 50;
var repScale = false;
var repScaleAmt = 95;
var groupFade = false;
var makeSymbols = false;// useful if processing complex images 

// dialogue values
var values = {
    spacing: spacing,
    spaceNo: spaceNo,
    spaceAdjust: spaceAdjust,
    fade:fade,
    rota: rota,
    vibraAm: vibraAm,
    rotationIncrement : rotationIncrement,
    rotaPx: rotaPx,
    rotaPy: rotaPy,
    mainScale: mainScale,
    jitter: false,
    jitterAmount:jitterAmount,
    grow: grow,
    growAmount: growAmount,
    repeatType: repeatType,   
    frontToBack: frontToBack,
    repeatGroup : repeatGroup,
    repeatGroupN: repeatGroupN,
    repScale: repScale,
    repScaleAmt: repScaleAmt,
    xOffset: xOffset,
    yOffset: yOffset,
    groupFade: groupFade,
    makeSymbols: makeSymbols 
};

var components = {
    spacing: { 
        type: 'list',
        label: 'Space type',
        options: [spaceTypeCount, spaceTypeSpace, spaceTypeFit, spaceTypeScaledFit, spaceTypeFitRand, spaceTypeScaledFitRand],
        onChange: function(value) {
		  if(value == spaceTypeFit || value == spaceTypeScaledFit || value == spaceTypeFitRand || value == spaceTypeScaledFitRand) {
			 components.spaceNo.enabled = false;
			 components.spaceAdjust.enabled = true;
		  } else {
			 components.spaceNo.enabled = true;
			 components.spaceAdjust.enabled = false;
		  }         
       }
    },
    spaceNo: { 
        type: 'number',
        label: 'Value',
        steppers: true
    },
	spaceAdjust: { 
	   type: 'number',
	   label: 'Adjust fit',
	   fractionDigits: 1,
	   units: 'percent',
	   enabled: false  // default is 'count' method
    },
    repeatType : {
       type: 'list',
       label: 'Cycle',
       options: [repeatTypeAll, repeatTypeRand, repeatTypeCycle]
    },
    fade : {
		type: 'boolean', 
		label: 'Fade'
    },
	ruler1: { //==================
		type: 'ruler'
    },
    rota: {
        type: 'list',
        label: 'Style',
        options: [rotaTypeOff, rotaTypeRotate, rotaTypeVib],
	    onChange: function(value) {
		  if (value == rotaTypeRotate) {
			 components.vibraAm.enabled = false;
			 components.rotationIncrement.enabled = true;
			 components.rotaPx.enabled = true;
			 components.rotaPy.enabled = true;
		  } else if (value == rotaTypeVib) {
			 components.vibraAm.enabled = true;
			 components.rotationIncrement.enabled = false;
			 components.rotaPx.enabled = true;
			 components.rotaPy.enabled = true;
		  } else { //off
			 components.vibraAm.enabled = false;
			 components.rotationIncrement.enabled = false;
			 components.rotaPx.enabled = false;
			 components.rotaPy.enabled = false;
		  }
        }
    },
    vibraAm: { 
        type: 'number',
        label: 'Vibrate',
        units: 'degree'
    },
    rotationIncrement: { 
        type: 'number',
        label: 'Rotation',
        units: 'degree',
        enabled: false //default is 'vibra'
    },
    rotaPx: { 
       type: 'number',
       label: 'Centre x',
       fractionDigits: 1,
       units: 'percent'
    },
    rotaPy: {
        type: 'number',
        label: 'Centre y',
        fractionDigits: 1,
        units: 'percent'
    },
    jitter: { 
		type: 'boolean', 
		label: 'Jitter'
	    },
	jitterAmount: { 
		type: 'number',
		label: 'J Amount',
		fractionDigits: 1,
		units: 'percent'
    },
    ruler2: { //==================
	     type: 'ruler'
    },	
    mainScale: {
		type: 'number',
		label: 'Main scale',
		fractionDigits: 1,
		units: 'percent'
    },    

    grow: { 
        type: 'list',
        label: 'Growth',
        options: [growUp, growDown, growOff],
    },  
	growAmount: {
		type: 'number',
		label: 'Scale',
		fractionDigits: 1,
		units: 'percent'
    },    
    frontToBack: { 
        type: 'boolean', 
        label: 'Front to back'
    },
    ruler3: { //==================
        type: 'ruler'
    },
    repeatGroup: { 
        type: 'boolean', 
        label: 'Repeat'
    },
    repeatGroupN: { 
        type: 'number',
        label: 'Count',
        steppers: true
    },
    xOffset: { 
        type: 'number',
        label: 'Offset x',
        steppers: true,
        units: 'pixel'
    },
    yOffset: { 
        type: 'number',
        label: 'Offset y',
        steppers: true,
        units: 'pixel'
    },
	repScale: { 
		type: 'boolean', 
		label: 'Scale'
    },      
    repScaleAmt: {
		type: 'number',
		label: 'Amount',
		fractionDigits: 1,
		units: 'percent'
	},        
	groupFade: { 
        type: 'boolean', 
        label: 'fade'
    },
    
    ruler4: { //=================
        type: 'ruler'
    },	
    makeSymbols: { 
        type: 'boolean', 
        label: 'Make symbols'
    },
    object: { 
        type: 'button',
        value:'Set',
        label: 'Objects',
        onClick: function() { SelectObjects(); }
    },
    path: { 
        type: 'button',
        value:'Set',
        label: 'Paths',
        onClick: function() {SelectPath();}
    },
    Run: { 
        type: 'button',
        value:'Run',
        label: 'Execute',
        onClick: function() {
            Run(); 
        }
    }
}  // dialog components

var palette = new Palette('Object on path', components, values);

// -----------------------------------------------------------------------------

function LoadParams() {
	// on Run, load parameters for efficiency ...
	spacing = values.spacing;
	spaceNo = values.spaceNo;
	spaceAdjust = values.spaceAdjust / 100;
	repeatType = values.repeatType;
	fade = values.fade;
	rota = values.rota;
	vibraAm = values.vibraAm;
	rotationIncrement = values.rotationIncrement;
	rotaPx = values.rotaPx / 100;
	rotaPy = values.rotaPy / 100;
	jitter = values.jitter;
	jitterAmount = values.jitterAmount / 100;
	repeatGroup = values.repeatGroup;
	repeatGroupN = values.repeatGroupN;
	xOffset = values.xOffset;
	yOffset = values.yOffset;
	repScale = values.repScale / 100;
	mainScale = values.mainScale / 100;
	repScaleAmt = values.repScaleAmt / 100;
	growAmount = values.growAmount / 100;
	grow = values.grow;
	if (grow == growOff) growAmount = 1;
	frontToBack = values.frontToBack;
	groupFade = values.groupFade;
	
} // fn LoadParams


function SelectPath() {
    pathList = document.getItems({
    type: Path,
        selected: true
    });
} //fn SelectPath

function SelectObjects() {

  objList = document.getItems({type: [Path,CompoundPath,Group,PlacedSymbol,TextItem], selected: true});
    if (! objList.length) {
        Dialog.alert('Select Objects!');
        return false;
    } 
  // Remove any children of another item so members of a group are not processed twice.      
  // Need to check for null in the main process
	for (obIdx = 0 ; obIdx < objList.length;  obIdx++) { 
		// check defined ( hasn't been nulled already )
		if (objList[obIdx]) {
		   obj = objList[obIdx];
		   for (obIdx2 = 0 ; obIdx2 < objList.length;  obIdx2++) {      
			   if (objList[obIdx2] && objList[obIdx2].isDescendant(obj)) {
				   objList[obIdx2] = null;
			   }    
		   } 
		} 
	}
	// turn the objects into placed symbols if required
	// note params are not loaded at this point
	if (values.makeSymbols) {
		for (obIdx = 0 ; obIdx < objList.length;  obIdx++) { 
			// check defined ( hasn't been nulled already ) and is not already a placed symbol
		   if (objList[obIdx] && objList[obIdx].type != 'PlacedSymbol') {
		      objList[obIdx] =  PlacedSymbol(objList[obIdx]);
		   }
		}
	}
}  // fn SelectObjects

function GetObjMaxSize() {
    // for 'fit' spacing need max size of the objects
    var size = 0;
	for (obIdx = 0; obIdx < objList.length; obIdx++) { 
		// check it is defined ( hasn't been nulled already )
		if (objList[obIdx]) {
		  var thisObj = objList[obIdx];
		  size = Math.max(size, thisObj.strokeBounds.width);
		  size = Math.max(size, thisObj.strokeBounds.height);
		}  
	} 
	return size;
}

// --------------------------------------------------------------------
function Run(){

    if (pathList == undefined || objList == undefined) {
        Dialog.alert('Select Object and Path before!');
        return false;
    } 
    
    // get the parameters 
    LoadParams()
    
    // group to contain all generated items for all paths
    var pathGroup = new Group(); 

    // in case needed for the 'fit' spacing 
    var objMaxSize = GetObjMaxSize() * spaceAdjust;

    // process all target paths 
	var pathCount = pathList.length;
	for ( var pathNo = 0 ; pathNo < pathCount; pathNo++) {
		
		var targetPath = pathList[pathNo];
		var pathLength = targetPath.length;		    
		
		var thisRotation = 0;      // changed for each position on path
	    var objScale = 1; // changed for each position

		// track which object to draw in 'cycle' mode - reset for each path
		var stepCycle = 0;
		
		// calc initial step size depending on spacing type
		var stepSize;
		var stepSizeVariation = 1.0;	
		switch (spacing) {
		case spaceTypeCount: 
		   stepSize = pathLength/spaceNo;
		   break;
		case spaceTypeSpace:
		   stepSize = spaceNo; 
		   break;
		case spaceTypeFit: 
		case spaceTypeScaledFit: 	
		   stepSize = objMaxSize * mainScale;
		   break;
		case spaceTypeFitRand: 
		case spaceTypeScaledFitRand:	
		   stepSize = objMaxSize * mainScale;
		   stepSizeVariation = (Math.random() * (ranFitMax - ranFitMin) + ranFitMin); // initial value - recalc in loop through points
        }
		
	    // process all positions on current path
		for (var pos = 0; pos <= pathLength; pos += stepSize * stepSizeVariation) {
		
	        if (stepSize < 0.1)  break;  // avoid zero sizes that make infinite loop

			// work from large object to small 
			var adjustedPos = pos;
			if (grow == growUp) { adjustedPos = pathLength - adjustedPos; } // work from the other end
			
			var point = targetPath.getPoint(adjustedPos);
			
			if (jitter) {
			   // calculate random jitter to offset point
			   var tangX = targetPath.getTangent(adjustedPos).x;
			   var tangY = targetPath.getTangent(adjustedPos).y;
			   var xyxy = Math.sqrt(tangX*tangX + tangY*tangY);
			   var jitterRan = objScale * mainScale * objMaxSize * jitterAmount * (Math.random() - 0.5) * 2 / xyxy;
			   point.x = point.x - jitterRan * tangY; // working in bottom up coordinates
			   point.y = point.y + jitterRan * tangX;
			}	
			
			function DrawObject(objIdx) { // -----------------------------------------------
			   // drawing an object - called from the three cycling methods
			   
			   // flags for matrix transform
			    var allFlagsOn = ['objects', 'fill-gradients', 'fill-patterns', 'stroke-patterns', 'lines', 'linked-masks', 'children'];
			    // keep selection-only flag OFF as it is the objects created are not selected

				// make new instance
				var objInstance = objList[objIdx].clone(); 
				
				// the scaling of this object
				var scaling = objScale * mainScale; 
				
				// set rotation on new instance at original position
				var mtx = new Matrix(1, 0, 0, 1, 0, 0);
				mtx = mtx.scale(scaling, objInstance.bounds.center);
				
				// rotate new instance	
				// calculate the point about which rotation is to occur
				var topLeft = objInstance.bounds.topLeft;
				var rotOffset = new Point(objInstance.bounds.width * rotaPx * scaling, -objInstance.bounds.height * rotaPy * scaling); 
				var rotationPoint = topLeft + rotOffset;
                
				if (rota == rotaTypeRotate) {
					mtx.rotate(thisRotation, rotationPoint);
					thisRotation += rotationIncrement;
				} else if (rota == rotaTypeVib) {
					mtx.rotate( (Math.random()*2 -1) * vibraAm, rotationPoint);
				}
				
				// apply scale and rotate 
				objInstance.transform(mtx, allFlagsOn);
				
				// position the object directly and not by translate
				objInstance.position = point;
				
				if (fade) {
				   objInstance.opacity = (Math.abs(pathLength- pos) / pathLength) * (1-fadeLimit) + fadeLimit ; 
				}

				// Add to the group
				// By default, bigger objects are at front
				if (frontToBack) {
				   pathGroup.appendBottom(objInstance);
				} else {
				   pathGroup.appendTop(objInstance);
				}

			} // fn drawObject -------------------------------
			
			
		   // process according to the repeat pattern
		   switch (repeatType) {
			case repeatTypeAll: 
			   for (var idx = 0; idx < objList.length; idx++) {
				  if (objList[idx]) { DrawObject(idx) } // skip nulls
			   }
			   break;
			case repeatTypeRand:
				do {
				  idx = Math.round( Math.random() * objList.length);
				} while (! objList[idx] );
				DrawObject(idx);
				break;
			case repeatTypeCycle:  
			     var done = false;
				 while (! done) {
				  for (var idx = stepCycle++; idx < objList.length && ! done; idx++) {
					 if (objList[idx]) { 
						 DrawObject(idx);
						 done = true;
					 }
				  } 
				  if (!done) stepCycle = 0; 
				}      
            } // switch-repeatpattern
          
			// Scale changes for next time 
			objScale = objScale * Math.abs(growAmount);	

			// Change spacing between objects for scaled fitting					
			if (spacing == spaceTypeScaledFit && pos != 0 ) {
			   stepSize = stepSize * Math.abs(growAmount); 
			}	 
			
			// randomise the fitting 
			if (spacing == spaceTypeFitRand  || spacing == spaceTypeScaledFitRand) {	
			   stepSizeVariation = (Math.random() * (ranFitMax - ranFitMin) + ranFitMin);
			}

	  } // points along the path
  
  } // process all paths
  
    // apply the repeat to the resulting pattern
    if (repeatGroup) RepeatGroup(pathGroup); // repeatGroup the generated  group
    
    // leave the paths selected so they can be deleted or changed for another run 
    document.deselectAll();
    pathCount = pathList.length;
	for (pathNo = 0 ; pathNo < pathCount; pathNo++) {
	   pathList[pathNo].selected = true;
	}
	     
} // fn Run


function RepeatGroup(group) {

    var groupScale = 1;
    
    total = new Group();
   
    total.appendTop(group); // the original
    
	for (j = 1; j < repeatGroupN; j++) {
	
		var tmp = group.clone();
		
		if (repScale) {
		   tmp.scale(groupScale);
		   tmp.translate(new Point(xOffset*j*groupScale, yOffset*j*groupScale));
		} else {  
		   tmp.translate(new Point(xOffset*j, yOffset*j));
		}
		
		if (groupFade) {
		   tmp.opacity = (repeatGroupN - j)/repeatGroupN * (1-fadeLimit) + fadeLimit ; 
		}
		
       // use the same f-to-b as objects for now .. may be better separate
		if (frontToBack) {
		   total.appendBottom(tmp);
		} else {
		   total.appendTop(tmp);
		}
		
		groupScale = groupScale * repScaleAmt;		
	 }
}// fn RepeatGroup